;***************************************************************************
; $Id: ic720.asm,v 1.1 2011/01/10 01:44:57 daveh Exp daveh $
;
;	ICOM IC-720A CI-V serial CAT interface program.
;	Connects RS-232 port to 4-bit parallel Icom interface.
;	Used 8-n-1 9600 baud protocol, no flow control.
;	Implements CAT commands 3-7 (minimally) at present.
;	Code based heavily on prototype code described below.
;	(Original header and documentation reproduced.)
;	See URL below for additional information, library code.
;	See also QST, pp. 34--38, July, 1984, pp. 30--33, May, 1981.	
;	Debugging simplified using interceptty package.
;
;
;	RS232 Test Interface V1.02
;	==========================
;
;	written by Peter Luethi, 26.03.1999, Dietikon, Switzerland
;	http://www.electronic-engineering.ch
;	last update: 16.01.2005
;
;	V1.02:	Fixed copy/paste issue of ISR context store/restore
;		(nobody is perfect): Erroneously erased INTCON,INTF
;		clearing, resulting in endless ISR calling...
;		Re-structured entire ISR and RS232 echo sub-routines
;		(11.04.2004)
;
;	V1.01:	ISR context restore improvements (21.11.1999)
;
;	V1.00:	Initial release (26.3.1999)
;
;	This code and accompanying files may be distributed freely and
;	modified, provided this header with my name and this notice remain
;	intact. Ownership rights remain with me.
;	You may not sell this software without my approval.
;
;	This software comes with no guarantee or warranty except for my
;	good intentions. By using this code you agree to indemnify me from
;	any liability that might arise from its use.
;
;
;	SPECIFICATIONS:
;	===============
;	Processor:			Microchip PIC 16F84A
;	Clock Frequency:		4.00 MHz XT
;	Throughput:			1 MIPS
;	Baud Rate:			depends on the module included
;	Code Size of entire Program:	approx. 235 instruction words
;	Required Hardware:		MAX 232
;
;
;	DESCRIPTION:
;	============
;	Developed and tested on PIC 16F84A, executeable on all interrupt
;	featured PICs.
;	Program handles all aspects of RS232
;		Transmission (Register TXD) and
;		Reception (register RXD) through interrupts (PortB0 IRQ).
;	The microcontroller sends feedback of received characters back to
;	the terminal window.
;
;	Program shows the implementation and function of the modules
;	m_bank.asm, m_wait.asm, and m_rs096.asm on the PIC16F84A.
;
;***************************************************************************

;***** COMPILATION MESSAGES & WARNINGS *****

	ERRORLEVEL -207 	; found label after column 1
	ERRORLEVEL -302 	; register in operand not in bank 0

;***** PROCESSOR DECLARATION & CONFIGURATION *****

	PROCESSOR 16F84A
	#include "p16f84a.inc"
	
	; embed Configuration Data within .asm File.
	__CONFIG   _CP_OFF & _WDT_OFF & _PWRTE_ON & _XT_OSC

;***** MEMORY STRUCTURE *****

	ORG     0x00			; processor reset vector
  	goto    MAIN

	ORG     0x04			; interrupt vector
	goto	ISR			; Interrupt Service Routine (ISR)

;***** PORT DECLARATION *****

	#define	TXport	PORTA,0x00	; RS232 output port, could be
	#define	TXtris	TRISA,0x00	; any active push/pull port

	#define DVport	PORTB,0x01	; Icom data valid
	#define RTport	PORTB,0x02	; Icom remote trigger	 	 
	#define DBCport	PORTB,0x03	; Icom data bus control

	#define DVtris	TRISB,0x01
	#define RTtris	TRISB,0x02
	#define DBCtris	TRISB,0x03
	; highest nibble of port B is for bidirectional data exchange
	
;***** CONSTANT DECLARATION *****

	CONSTANT BASE = 0x0C		; base address of user file regs 
	CONSTANT BUFBASE = 0x22		; base address for serial buffer 
	CONSTANT BUFLEN = 0x10		; serial buffer max length 
	
;***** REGISTER DECLARATION *****

	CBLOCK 0x0c
	
	FLAGreg			; used for many things - take care
	
	TXD	; TX-Data register
	RXD	; RX-Data register

	W_TEMP	; context register (ISR)
	STATUS_TEMP ; context register (ISR)
	PCLATH_TEMP ; context register (ISR)
	FSR_TEMP ; context register (ISR)

	WCYCLE
	TEMP1
	TEMP2
	TEMP3

	FROM			; CAT address of PC
	F0			; Icom frequency byte 0 LSB
	F1			; Data stored in BCD format
	F2			; Little endian, F0 not really used
	F3			; Icom frequency byte 3 MSB
	F3P			; previous value of F3
	MODE			; Icom mode
	VFO			; Icom VFO a/b
	PTR			; pointer to serial buffer
	MSG			; message word for serial reply
	
	ENDC

	#define	RSflag	FLAGreg,0x00	; RS232 data reception flag
	#define RDflag	FLAGreg,0x04	; packet ready flag

	#define TO 0x04		; IC735 CAT address used here
	
;***** INCLUDE FILES *****

	#include "m_bank.asm"
	#include "m_wait.asm"
	#include "m_rs096.asm"	; specify desired RS232 module

;***** MACROS *****

;***** SUB-ROUTINES *****

	; 4-bit Icom port subroutines here

TableLook1			; convert CI-V mode format to hardware format
	addwf	PCL,f
	dt 0x0b, 0x00, 0x08, 0x06, 0x0c

TableLook2
	addwf	PCL,f		; convert hardware mode format to CI-V format
	dt 0x01,0x00,0x00,0x00,0x00,0x00,0x03,0x00,0x02,0x00,0x00,0x00,0x04
	
Hs				; start Icom port handshake
loop0	btfsc	DVport		; DV should be low prior to handshake
	goto	loop0
	bsf	RTport		; assert trigger
loop1	btfss	DVport		; wait for DV to go high
	goto	loop1		; need a watchdog?
	WAIT	d'1'		; read settle time
	RETURN

PHN				; put high nibble of w into 4-bit port
	andlw	0xf0		; destructive operation
	iorwf	PORTB,f
	bsf	DBCport		; assert output
	call	Hs
	bcf	RTport		; handshake done
	bcf	DBCport		; drop output
	movlw	0x0f
	andwf	PORTB,f		; clear port
	RETURN
	
PLN				; put low nibble of w into 4-bit port
	andlw	0x0f		; destructive operation
	clrc
	movwf	TEMP3		
	rlf	TEMP3,f
	rlf	TEMP3,f
	rlf	TEMP3,f
	rlf	TEMP3,f
	movfw	TEMP3
	iorwf	PORTB,f
	bsf	DBCport		; assert output
	call	Hs
	bcf	RTport		; handshake done
	bcf	DBCport		; drop output
	movlw	0x0f
	andwf	PORTB,f		; clear port
	RETURN	
	
GHN				; get port data into high nibble of w
	call	Hs
	movfw	PORTB
	bcf	RTport		; handshake done
	andlw	0xf0
	RETURN

GLN				; get port data into low nibble of w
	call	Hs
	movfw	PORTB
	bcf	RTport		; handshake done
	andlw	0xf0
	clrc
	movwf	TEMP3
	rrf	TEMP3,f
	rrf	TEMP3,f
	rrf	TEMP3,f
	rrf	TEMP3,f
	movfw	TEMP3	
	RETURN
	
ICwrite				
	call	ICid		; send the Icom ID (leaves port readable)

	BANK1
	bcf	TRISB,7		; port set for writing
	bcf	TRISB,6
	bcf	TRISB,5
	bcf	TRISB,4
	BANK0
	
	movfw	MODE		; put mode
	call	TableLook1	; convert to hardware format
	call	PLN
	
	movfw	VFO		; put VFO
	iorlw	0x0a		; add bits to form hardware address
	call	PLN	
	
	movfw	F3		; put F3
	call	PHN
	movfw	F3
	call	PLN
	
	movfw	F2		; put F2
	call	PHN
	movfw	F2
	call	PLN
	
	movfw	F1		; put F1
	call	PHN
	movfw	F1
	call	PLN

	BANK1			; put port back into read mode
	bsf	TRISB,7
	bsf	TRISB,6
	bsf	TRISB,5
	bsf	TRISB,4
	BANK0

	movfw	F3P
	subwf	F3,w
	bz	Unch
	WAIT	0xff		; if MSB changes, sleep for 2 s
	WAIT	0xff		; to allow possible band change
	WAIT	0xff
	WAIT	0xff
	WAIT	0xff
	WAIT	0xff
	WAIT	0xff
	WAIT	0xff
	movfw	F3
	movwf	F3P
	
Unch	RETURN

ICread
	clrf	F0		; clear all Icom registers
	clrf	F1
	clrf	F2
	clrf	F3
	clrf	MODE
	clrf	VFO
	
	call	ICid		; send the Icom ID (leaves port readable)

	call	GLN		; get MODE
	call	TableLook2	; convert from hardware format
	iorwf	MODE,f

	call	GLN		; get VFO
	andlw	0x01		; mask unused bits
	iorwf	VFO,f
	
	call	GHN		; get F3
	movwf	F3
	call	GLN
	iorwf	F3,f

	call	GHN		; get F2
	movwf	F2
	call	GLN
	iorwf	F2,f

	call	GHN		; get F1
	movwf	F1
	call	GLN
	iorwf	F1,f

	movfw	F3
	movwf	F3P		
	
	RETURN

ICid				; write IC-720A ID, no handshaking

	BANK1
	bcf	TRISB,7		; port set for writing
	bcf	TRISB,6
	bcf	TRISB,5
	bcf	TRISB,4
	BANK0
	
	movlw	0xf0
	iorwf	PORTB,f
	bcf	PORTB,4		; program port B with Icom 720 ID 0x0e
	bsf	DBCport		; assert output	
	WAIT	d'50'		; wait 50 ms
	bcf	DBCport		; drop output	

	WAIT	d'10'
	bsf	RTport		; assert trigger	
	WAIT	d'2'
	bcf	RTport		; clear trigger
	
	movlw	0x0f
	andwf	PORTB,f		; clear port
	
	BANK1			; put port back into read mode
	bsf	TRISB,7
	bsf	TRISB,6
	bsf	TRISB,5
	bsf	TRISB,4
	BANK0
			
	RETURN

	; serial port messaging subroutines

EchoM	movwf	MSG		;  sends echo of received command
	movlw	0xfe		;  header info follows
	SENDw
	movlw	0xfe
	SENDw
	movlw	TO
	SENDw
	movfw	FROM
	SENDw
	movfw	MSG
	SENDw			;  done with boilerplate
	movfw	MSG
	addlw	0xfb
	bz	FreqC		;  include freq data in echo
	addlw	0xff
	bz	ModeC		;  include mode in echo
	addlw   0xff
	bz	VFOC		;  include VFO in echo
	goto	Term
	
SendM	movwf	MSG		;  send a reply message 
	movlw	0xfe		;  header info follows
	SENDw
	movlw	0xfe
	SENDw
	movfw	FROM
	SENDw
	movlw	TO
	SENDw
	movfw	MSG
	SENDw			;  done with boilerplate
	movfw	MSG
	addlw	0xfd
	bz	FreqC		;  include freq data in reply
	addlw	0xff
	bz	ModeC		;  include  mode in reply
Term	movlw	0xfd		;  terminate character
	SENDw
	RETURN

FreqC	movfw	F0
	SENDw
	movfw	F1
	SENDw
	movfw	F2
	SENDw
	movfw	F3
	SENDw
	goto	Term

ModeC	movfw	MODE
	SENDw
	goto	Term

VFOC	movfw	VFO
	SENDw
	goto	Term
	
	; serial port polling routines
		
RSservice
	movfw	PTR		; get buffer pointer
	movwf	FSR		; put on FSR
	movfw	RXD		; get input character
	movwf	INDF		; add to buffer
	sublw	0xfe		; test for packet start
	bz	RSreset		; reset buffer pointer if so
	incf	FSR,f		; increment buffer pointer
	movfw	FSR		; 
	movwf	PTR		; save it
	sublw	BUFBASE+BUFLEN	; check for overflow
	bnz	RScont
RSreset
	movlw	BUFBASE		; if so, reset buffer pointer
	movwf	PTR		; to base address
RScont
	movfw	RXD		; re-examine input
	sublw	0xfd		; test for packet end
	bnz	RSend
	bsf	RDflag		; set packet ready flag if so
	RETURN			; and return w/o enabling interruots 
RSend
	; used to echo here
	; end of RS232 service (echo & display)
	bcf	RSflag		; reset RS232 data reception flag
	bsf	INTCON,INTE	; re-enable RB0/INT interrupt
	RETURN

;***** INTERRUPT SERVICE ROUTINE *****

ISR	;************************
	;*** ISR CONTEXT SAVE ***
	;************************

	bcf	INTCON,GIE	; disable all interrupts
	btfsc	INTCON,GIE	; assure interrupts are disabled
	goto	ISR
	movwf	W_TEMP		; context save: W
	swapf	STATUS,W	; context save: STATUS
	movwf	STATUS_TEMP	; context save
	clrf	STATUS		; bank 0, regardless of current bank
	movfw	PCLATH		; context save: PCLATH
	movwf	PCLATH_TEMP	; context save
	clrf	PCLATH		; page zero, regardless of current page
	bcf	STATUS,IRP	; return to bank 0
	movfw	FSR		; context save: FSR
	movwf	FSR_TEMP	; context save
	;*** context save done ***

	;**************************
	;*** ISR MAIN EXECUTION ***
	;**************************
	
	;*** check origin of interrupt ***
	btfsc	INTCON,INTF	; check for RB0/INT interrupt
	goto	_ISR_RS232	; if set, there was a keypad stroke

	; catch-all
	goto	ISRend		; unexpected IRQ, terminate execution of ISR

	;******************************
	;*** RS232 DATA ACQUISITION ***
	;******************************
_ISR_RS232
	; first, disable interrupt source
	bcf	INTCON,INTE	; disable RB0/INT interrupt
	; second, acquire RS232 data
	RECEIVE			; macro of RS232 software reception
	bsf	RSflag		; enable RS232 data reception flag
	goto	_ISR_RS232end	; terminate RS232 ISR properly

	;***********************************
	;*** CLEARING OF INTERRUPT FLAGS ***
	;***********************************
	; NOTE: Below, I only clear the interrupt flags! This does not
	; necessarily mean, that the interrupts are already re-enabled.
	; Basically, interrupt re-enabling is carried out at the end of
	; the corresponding service routine in normal operation mode.
	; The flag responsible for the current ISR call has to be cleared
	; to prevent recursive ISR calls. Other interrupt flags, activated
	; during execution of this ISR, will immediately be served upon
	; termination of the current ISR run.
_ISR_RS232error
	bsf	INTCON,INTE	; after error, re-enable IRQ already here
_ISR_RS232end
	bcf	INTCON,INTF	; clear RB0/INT interrupt flag
	;goto	ISRend		; terminate execution of ISR

	;*****************************************
	;*** ISR TERMINATION (CONTEXT RESTORE) ***
	;*****************************************

ISRend	movfw	FSR_TEMP	; context restore
	movwf	FSR		; context restore
	movfw	PCLATH_TEMP	; context restore
	movwf	PCLATH		; context restore
	swapf	STATUS_TEMP,W	; context restore
	movwf	STATUS		; context restore
	swapf	W_TEMP,F	; context restore
	swapf	W_TEMP,W	; context restore
	RETFIE			; enable global interrupt (INTCON,GIE)

;***** END OF INTERRUPT SERVICE ROUTINE *****


;************** MAIN **************

MAIN	
	clrf	INTCON		; reset interrupts (disable all)	
	RS232init		; RS232 initialization
	clrf	FLAGreg		; initialize all flags

	movlw	BUFBASE		; get buffer base
	movwf	PTR		; set buffer pointer
	bcf	RDflag		; clear serial packet ready flag
	
	SEND	CR		; Carriage Return
	SEND	LF		; Line Feed
	SEND	'I'
	SEND	'C'
	SEND	'-'
	SEND	'7'
	SEND	'2'
	SEND	'0'
	SEND	'A'
	SEND	' '
	SEND	'c'
	SEND	'a'
	SEND	't'
	SEND	'.'
	SEND	'.'
	SEND	'.'
	SEND	CR		; Carriage Return
	SEND	LF		; Line Feed
	SEND	LF		; Line Feed

	BANK1
	bcf	RTtris		; RT is an output
	bsf	DVtris		; DV is an input
	bcf	DBCtris		; DBC is an output

	bsf	TRISB,7		; 4-bit port initially set for read
	bsf	TRISB,6
	bsf	TRISB,5
	bsf	TRISB,4
	BANK0

	movlw	d'6'
	movwf	TEMP3
	;; call	ICid		; clear buffer to initialize
	;; WAIT	d'2'
CNT	bsf	RTport		; assert trigger	
	WAIT	d'2'
	bcf	RTport		; clear trigger
	WAIT	d'2'
	decfsz	TEMP3,1
	goto	CNT	
	
	;; call	ICread		; load registers with data from radio
	
LOOP	btfsc	RSflag		; check RS232 data reception flag
	call	RSservice	; if set, call RS232 handler
	btfss	RDflag		; test for complete packet
	goto	LOOP

	movlw	BUFBASE+d'1'	; capture FROM address
	movwf	FSR
	movfw	INDF
	movwf	FROM		; store it in register
	
	movlw	BUFBASE+d'2'	; index of input command
	movwf	FSR		; identify command
	movfw	INDF		; and branch
	addlw	0xfd		; read freq command 3
	bz	ICRfreq	
	addlw	0xff		; read mode command 4
	bz	ICRmode
	addlw	0xff		; write freq command 5
	bz	ICWfreq
	addlw	0xff		; write mode command 6
	bz	ICWmode
	addlw	0xff		; write VFO command 7
	bz	ICWvfo		

	movlw	0xfa		; send generic failure message
	call	SendM		
	; bad packet? restart serial system
	
RESTART
	movlw	BUFBASE		; set buffer pointer
	movwf	PTR		; to base address
	bcf	RDflag		; clear packet ready flag	
	
	bcf	RSflag		; reset RS232 data reception flag
	bsf	INTCON,INTE	; re-enable RB0/INT interrupt	
	goto	LOOP

ICRfreq				; read frequency from radio
	call	ICread		; communicate with radio

	movlw	0x03
	call	EchoM
	
	movlw	0x03
	call	SendM		; extended reply to PC over RS232
		
	goto	RESTART

ICRmode				; read mode from radio

	call	ICread		; communicate with radio

	movlw	0x04
	call	EchoM
	
	movlw	0x04
	call	SendM		; extended reply to PC over RS232
				
	goto	RESTART
	
ICWmode				; write mode to radio
	movlw	BUFBASE+d'3'	; get pointer to mode being set	
	movwf	FSR		; put on FSR
	movfw	INDF
	movwf	MODE		; set mode register

	call	ICwrite		; communicate with radio

	movlw	0x06
	call	EchoM	
	
	movlw	0xfb
	call	SendM		; generic reply to PC over RS232
	
	goto	RESTART

ICWvfo				; write vfo to radio (a/b)
	movlw	BUFBASE+d'3'	; get pointer to vfo being set	
	movwf	FSR		; put on FSR
	movfw	INDF
	movwf	VFO		; set vfo register

	call	ICwrite		; communicate with radio

	movlw	0x07
	call	EchoM
	
	movlw	0xfb
	call	SendM		; generic reply to PC over RS232
	goto	RESTART
	
ICWfreq				; write frequency to radio
	movlw	BUFBASE+d'3'	; get pointer to freq being set	
	movwf	FSR		; put on FSR
	movfw	INDF
	movwf	F0		; byte 0
	incf	FSR,f
	movfw	INDF
	movwf	F1
	incf	FSR,f
	movfw	INDF
	movwf	F2
	incf	FSR,f
	movfw	INDF
	movwf	F3		; byte 3 

	call	ICwrite		; communicate with radio

	movlw	0x05
	call	EchoM
	
	movlw	0xfb
	call	SendM		; generic reply to PC over RS232
	goto	RESTART
	
	END
